/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 */

#include "stdafx.h"

char input_rcsid[] = 
  "$Id: input.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


/*
 * Routing Table Management Daemon
 */

#include "defs.h"

extern int filter;
struct timeval now;		/* current idea of time */
int changes = 0;                /* is set if changes occur in MPRS set */ 

void tools_input(struct sockaddr *from, struct afswitch *afp, 
		 struct olsr *olsr, int size);

/*
 * Build HELLO Message according to hello_message struct from olsrmsg struct.
 */

void hello_chgestruct(struct hello_message *hmsg, struct olsrmsg *m,
                      struct sockaddr *from)
{
 
        struct hellomsg *h;
	struct hellinfo *hinfo, *hinf;
        olsr_ip_addr *haddr, *hadr;
        struct hello_neighbor *nb;
  
        hmsg->neighbors = (struct hello_neighbor *)NULL;
      
        if (!m)
	    return;

        if (m->olsr_msgtype != HELLO_PACKET && 
	    m->olsr_msgtype != UPDATE_NEIGHBOR)
	    return;

        h = m->olsr_hello;
        hinfo = h->hell_info;

        hmsg->source_addr = ((struct sockaddr_in *)from)->sin_addr.s_addr;
        hmsg->packet_seq_number = ntohs(h->hell_seqnum);
        hmsg->mpr_seq_number = ntohs(h->hell_mprseqnum);

        for (hinf = hinfo; (char *)hinf < (char *)m 
                           + ntohs(m->olsr_nextmsg); 
             hinf = (struct hellinfo *)((char *)hinf 
             + ntohs(hinf->next_linktype))
            ){

            haddr = hinf->neigh_addr;     

            for (hadr = haddr; (char *)hadr < (char *)hinf 
                 + ntohs(hinf->next_linktype); hadr++
                ){
                
       	        nb = (struct hello_neighbor *)malloc(
                       sizeof (struct hello_neighbor));
                if (nb == 0){
		  printf("olsrd: out of memory \r\n");
                  break;
		}
	        nb->address = *hadr;
                nb->status = hinf->link_type;
                nb->next = hmsg->neighbors;
                hmsg->neighbors = nb;
            }
       }
}


/*
 * Build TC Message according to tc_message struct from olsrmsg struct.
 */

void tc_chgestruct(struct tc_message *tmsg, struct olsrmsg *m,
                      struct sockaddr *from)
{

        struct tcmsg *tc;
        olsr_ip_addr *mprsaddr, *maddr;
	struct tc_mpr_addr *mprs;

        tmsg->multipoint_relay_selector_address = 
                              (struct tc_mpr_addr *)NULL;

        if (!m)
	    return;

        if (m->olsr_msgtype != TC_PACKET)
	    return;

        tc = m->olsr_tc;
        mprsaddr = tc->tc_mprsaddr;

        tmsg->source_addr = ((struct sockaddr_in *)from)->sin_addr.s_addr;
        tmsg->originator = tc->tc_origaddr;
        tmsg->packet_seq_number = ntohs(tc->tc_seqnum);
        tmsg->hop_count =  tc->tc_hopcnt;
        tmsg->mssn =  ntohs(tc->tc_mprseqnum);

        for (maddr = mprsaddr; (char *)maddr < (char *)m 
             + ntohs(m->olsr_nextmsg); 
             maddr++
            ){

             mprs = (struct tc_mpr_addr *)malloc(sizeof(struct tc_mpr_addr));
             if (mprs == 0){
		  printf("olsrd: out of memery \r\n");
                  break;
		}
	     mprs->address = *maddr;
             mprs->next = tmsg->multipoint_relay_selector_address;
             tmsg->multipoint_relay_selector_address = mprs;

        }

}


/*
 * Process a newly received packet.
 */

void olsr_input(struct sockaddr *from, struct olsr *olsr, int size)
{
	struct olsrmsg *m = olsr->olsr_msg;
    struct hello_message hellopacket;
    struct tc_message tcpacket_in, *tcpacket_out;
    struct rt_message *routepacket = NULL;
	olsr_ip_addr src_addr = 
	  ((struct sockaddr_in *)from)->sin_addr.s_addr;
	struct _interface *ifp;
    int minsize;      /* minimum size of olsr packet */
	int count;
	struct afswitch *afp;
	char buf1[256];


        tcpacket_in.multipoint_relay_selector_address = 
                              (struct tc_mpr_addr *)NULL;
        hellopacket.neighbors = (struct hello_neighbor *)NULL;

        afp = &afswitch[from->sa_family];

	if (from->sa_family >= af_max || afp->af_hash == (void (*)(struct sockaddr *, struct afhash *))0) {
/*		syslog(LOG_INFO,
	              "\"from\" address in unsupported address family (%d),
		      msg %d\n",from->sa_family, m->olsr_msgtype);*/
	    sprintf(logdata,"\"from\" address in unsupported address family (%d), msg %d\n",from->sa_family, m->olsr_msgtype);
	    LOG(logdata);
		    
	    return;
	}
	
	     
	if (ntohs(olsr->olsr_packlen) != size) {

            printf("Size error detected in received packet.\r\n");

/*            syslog(LOG_ERR,
	           " packet length error in  packet received from %s!",
	          (*afswitch[from->sa_family].af_format)(from, buf1,
	    					   sizeof(buf1)));
*/

	    sprintf(logdata," packet length error in  packet received from %s!", (*afswitch[from->sa_family].af_format)(from, buf1, sizeof(buf1)));
	    LOG(logdata);
	    return;
	}

       /*
	* Process tools packets.
	*/


	tools_input(from, afp, olsr, size);
	
       /* verify message came from a privileged port */
       if ((*afp->af_portmatch)(from) == 0)
	     	return;

       (*afp->af_canon)(from);

       /* are we talking to ourselves? */
       ifp = if_ifwithaddr(from);
       if (ifp) {
	  if (ifp->int_flags & IFF_PASSIVE) {

             printf("Data from passive interface\r\n");
/*	     syslog(LOG_ERR,
		   "bogus input (from passive interface, %s)",
		   (*afswitch[from->sa_family].af_format)(from,
		   buf1, sizeof(buf1)));
*/
	     sprintf(logdata, "bogus input (from passive interface, %s)",
		   (*afswitch[from->sa_family].af_format)(from,
		   buf1, sizeof(buf1)));
	     LOG(logdata);
	     return;
	  }

         return;
       }


      if (filter)
	 if (!is_neighbor(&addr, (struct sockaddr_in *)from))
	     return;

	ifp = 0;

	TRACE_INPUT(ifp, from, (char *)olsr, size);

        minsize = (int)sizeof(m->olsr_msgtype) + 
	          (int)sizeof(m->olsr_broad) +
                  (int)sizeof(m->olsr_nextmsg);

       count = size - ((char *)m - (char *)olsr);

       if (count < minsize)
               return;
       for ( ; count > 0; m = (struct olsrmsg *)((char *)m + ntohs(m->olsr_nextmsg)) ){

	   if (count < minsize)
	              break;
	   count -= ntohs(m->olsr_nextmsg);

           switch (m->olsr_msgtype) {

           case HELLO_PACKET:
          
			hello_chgestruct(&hellopacket, m, from);
		
			olsr_process_received_hello(&hellopacket);

			break;

           case TC_PACKET:

             tc_chgestruct(&tcpacket_in, m, from);

			 olsr_process_received_tc(&tcpacket_in, &tcpacket_out);
             
             if (tcpacket_out)
				toall(tc_build, (struct hello_message *)NULL, 
					tcpacket_out, (struct _interface *)NULL);
	     
           break;

           default:

	   break;

	   } /* switch */

       } /* for olsr_msg */ 

//    sigpending(&psigset);
    
    /* If SIGALRM is pending, wait for the end of   
     * the received packet process to send the stored
     * msg(s) with next broadcast msg(s).
     */

       // if ( outputsize && !sigismember(&psigset, SIGALRM) )
       if (outputsize)
       toall(sndmsg, (struct hello_message *)NULL, 
                        (struct tc_message *)NULL, 
		        (struct _interface *)NULL );

}



       /*
	* Process tools packets.
	*/


void tools_input(struct sockaddr *from, struct afswitch *afp, 
		 struct olsr *olsr, int size)
{

	struct olsrmsg *m = olsr->olsr_msg;
        struct hello_message hellopacket;
        struct rt_message *routepacket = NULL;
	struct _interface *ifp = 0;

        switch (m->olsr_msgtype) 
	{

           case HELLO_REQUEST:

	     olsr_build_hello_packet(&hellopacket);

	     supply(from, &hellopacket, (struct rt_message *)NULL, 
                          m->olsr_msgtype);

	     return;

           case ROUTE_REQUEST:

	     olsr_build_routing_packet(&routepacket);

	     supply(from, (struct hello_message *)NULL, routepacket,
                           m->olsr_msgtype);

	     return;

           case UPDATE_NEIGHBOR:

	     hello_chgestruct(&hellopacket, m, from);

	     update_static_neighborhood(&hellopacket);
	     
	     return;

	   case OLSRD_TRACEON:
	   case OLSRD_TRACEOFF:

	    /* verify message came from a privileged port */

	     if ((*afp->af_portcheck)(from) == 0)
	       return;
 
	      ((char *)olsr)[size] = '\0';
	       if (m->olsr_msgtype == OLSRD_TRACEON)
		 {
	        	traceon(m->olsr_tracefile);
		        bumploglevel();
		        bumploglevel();
		        bumploglevel();
		 }
	       else
			traceoff();
	       return;
        }


}
