/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 *	$Id: neighbor_table.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */

/*
 * Routing table management daemon.
 */


/*
 *Neighbor table structure.
 */

struct neighbor_2_list_entry {
	struct	neighbor_2_list_entry *neighbor_2_next;
	struct	neighbor_2_entry      *neighbor_2;
        struct  timeval	              neighbor_2_timer;
};


struct neighbor_info {

	olsr_u32_t                   neighbor_hash;
	olsr_ip_addr                 neighbor_addr;
	olsr_u8_t                    neighbor_status;
        struct timeval               neighbor_timer;
        int                          neighbor_2_nocov;
      	struct _interface             *neighbor_ifp;	      
	struct neighbor_2_list_entry *neighbor_2_list; 
};


struct neighbor_entry {
	struct	neighbor_entry *neighbor_forw;
	struct	neighbor_entry *neighbor_back;
        struct  neighbor_info neighbor_infos;
};


#define	neighbor_hash	    neighbor_infos.neighbor_hash
#define	neighbor_addr	    neighbor_infos.neighbor_addr                             /* one-hop neighbor address */
#define	neighbor_status     neighbor_infos.neighbor_status                           /* status of the link between 
							                          the neighbor and this node (see bellow) */
#define	neighbor_2_nocov    neighbor_infos.neighbor_2_nocov                    /* number of neighbors to the one
							                          hop neighbor and not yet covered */
#define	neighbor_timer      neighbor_infos.neighbor_timer                      /* for invalidation */
#define	neighbor_ifp        neighbor_infos.neighbor_ifp                        /* interface to take to this neighbor */
#define	neighbor_2_list     neighbor_infos.neighbor_2_list                     /* list of two hops neighbors
							                          accessible by this one hop neighbor */

struct neighborhash {
	struct	neighbor_entry *neighbor_forw;
	struct	neighbor_entry *neighbor_back;
};

struct neighbor_table {
       olsr_u16_t          neighbor_mpr_seq;
       struct neighborhash *neighborhash;
};


/*
 * "Status" of link.
 */
#define	NBS_ASYM	1		/* uni-directional link */
#define	NBS_SYM 	2		/* bi-directional link */
#define	NBS_MPR 	3		/* neighbor has been selected
					   as MultiPoint Relay */
#define NBS_LOST        4               /* Loosing a link */   

#define	HELLOSTATUS_MAX		4


int 
olsr_delete_neighbor_2_pointer(struct neighbor_entry *neighbor,olsr_ip_addr *address);

struct neighbor_2_list_entry
*olsr_lookup_my_neighbors(struct neighbor_entry *neighbor,olsr_ip_addr *neighbor_address);

int
olsr_delete_neighbor_table(struct neighbor_entry *neighbor);

int
olsr_insert_neighbor_table(struct neighbor_entry *neighbor);

struct neighbor_entry
*olsr_lookup_neighbor_table(olsr_ip_addr *dst);

int
olsr_release_neighbor_table();

int
olsr_time_out_two_hop_neighbors(struct neighbor_entry  *neighbor);

int
olsr_time_out_neighborhood_tables();

int
olsr_print_neighbor_table();
