/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

#include "stdafx.h"

char olsr_rcsid[] = 
  "$Id: olsr.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


#include "defs.h"

/* --------------------------------------------------------------------
    OLSR protocol
   -------------------------------------------------------------------- */

/* --------------------------------------------------------------------
    Forward declarations
   -------------------------------------------------------------------- */


/* --------------------------------------------------------------------
    Global variables and arrays
   -------------------------------------------------------------------- */


/* --------------------------------------------------------------------
    Functions
   -------------------------------------------------------------------- */


/*-------------------------------------------------------------------------*/

struct timeval  hold_time_neighbor ;
struct timeval  hold_time_topology;
struct timeval  hold_time_duplicate_tc;

olsr_8_t changes_topology;
olsr_8_t changes_neighborhood;

olsr_ip_addr default_gateway=0;
/*--------------------------------------------------------------------------*/
FILE *f_r_id;

/*-------------------------------------------------------------------------*/
int
olsr_hashing(olsr_ip_addr *address,olsr_u32_t *h)
{
  olsr_u32_t hash;
  
  hash = (ntohl(*address));
  hash &= 0x7fffffff; 
  *h=hash;

	return 0;
}
/*-------------------------------------------------------------------------*/

struct neighbor_2_list_entry 
*olsr_find_2_hop_neighbors_with_1_link()

{
  
 
  olsr_u8_t                    index;
  struct neighbor_2_list_entry *two_hop_list_tmp=NULL;
  struct neighbor_2_list_entry *two_hop_list=NULL;
  struct neighbor_2_entry      *two_hop_neighbor=NULL;
  struct neighbor2_hash        *hash_2_neighbor;


  for(index=0;index<HASHSIZE;index++)
    {
      hash_2_neighbor=&neighbor2table[index];

      for(two_hop_neighbor=hash_2_neighbor->neighbor2_forw;two_hop_neighbor!=(struct neighbor_2_entry  *)hash_2_neighbor;two_hop_neighbor=two_hop_neighbor->neighbor_2_forw)
	{

	  two_hop_neighbor->neighbor_2_state=0;

	  if(two_hop_neighbor->neighbor_2_pointer==1)
	    {
	      if(two_hop_neighbor->neighbor_2_nblist->neighbor->neighbor_status==NBS_SYM)
		{
		  two_hop_list_tmp =(struct neighbor_2_list_entry *)malloc(sizeof(struct neighbor_2_list_entry));
		  two_hop_list_tmp->neighbor_2=two_hop_neighbor;
		  two_hop_list_tmp->neighbor_2_next=two_hop_list;
		  two_hop_list= two_hop_list_tmp;
		}
	    }
	  
	}
      
    }
  
  return(two_hop_list_tmp);
}
  
/*---------------------------------------------------------------------------*/

int
olsr_chosen_mpr(struct neighbor_entry *one_hop_neighbor,olsr_u16_t *two_hop_covered_count)
{
  /*
    This function  processes the chosen  and updates the counters
    used in calculations  
  */
  
  struct neighbor_list_entry   *the_one_hop_list;
  
  struct neighbor_2_list_entry *second_hop_entries; 
  olsr_u16_t                   count=0;
  
  count=*two_hop_covered_count;
  
  one_hop_neighbor->neighbor_status=NBS_MPR;

  TRACE_ACTION("CHANGE NEIGHBOR STATUS : MPR ",&one_hop_neighbor->neighbor_addr,NULL,0);

  second_hop_entries=one_hop_neighbor->neighbor_2_list;
  
  while(second_hop_entries!=NULL)
    
    {
      if(!second_hop_entries->neighbor_2->neighbor_2_state)
	{
	  /*
	    Now the neighbor is covered by this mpr
	  */
	  second_hop_entries->neighbor_2->neighbor_2_state=NB2S_COVERED;  /* =1*/	  
	  
	  the_one_hop_list=second_hop_entries->neighbor_2->neighbor_2_nblist;
	  
	  while(the_one_hop_list!=NULL)
	    {
	     
	      if((the_one_hop_list->neighbor->neighbor_status!=NBS_ASYM)&&(the_one_hop_list->neighbor->neighbor_status!=NBS_LOST))
		{
		  count++;
		  the_one_hop_list->neighbor->neighbor_2_nocov--;
		}
	      the_one_hop_list=the_one_hop_list->neighbor_next;
	      
	    }
	  
	}
      second_hop_entries=second_hop_entries->neighbor_2_next;
    }
  
  *two_hop_covered_count=count;
return (0);

}
/*--------------------------------------------------------------------------*/
int
olsr_calculate_neighbors(olsr_u16_t *two_hop_count)
{
  /*
    This function calculates the number of  second hop neighbors for each 
    neighbor
  */


  olsr_u8_t                    index;
  struct neighborhash          *neighbor_hash_table;
  struct neighbor_entry        *a_neighbor;
  olsr_u16_t                   count=0;
  struct neighbor_2_list_entry *twohop_neighbors;
  
  
  neighbor_hash_table=neighbortable.neighborhash;
  
  for (index=0;index<HASHSIZE;index++)
    
    {
      for(a_neighbor=neighbor_hash_table[index].neighbor_forw;a_neighbor!=(struct neighbor_entry  *)&neighbor_hash_table[index];a_neighbor=a_neighbor->neighbor_forw)
	{ 
	  
	  count=0;

	  if((a_neighbor->neighbor_status==NBS_ASYM)||(a_neighbor->neighbor_status==NBS_LOST))
	    {	    
	      a_neighbor->neighbor_2_nocov=count;
	      continue;
	    }
	  /* 
	     Turn the neighbor status on symetric to re calculate the MPR set
	  */
	  if(a_neighbor->neighbor_status==NBS_MPR)
	    a_neighbor->neighbor_status=NBS_SYM;
	  count=0;
	  twohop_neighbors=a_neighbor->neighbor_2_list;
	  while(twohop_neighbors!=NULL)
	    {
	      count++;
	      twohop_neighbors=twohop_neighbors->neighbor_2_next;
	    }
	  a_neighbor->neighbor_2_nocov=count;
	  *two_hop_count= *two_hop_count+count;
	} 
    }

	return 0;
}
/*--------------------------------------------------------------------------*/
struct neighbor_entry 
*olsr_find_maximum_covered()
{
  olsr_u16_t                  maximum=0;
  olsr_u8_t                   index;
  struct neighborhash         *neighbor_hash_table;
  struct neighbor_entry       *a_neighbor;
  struct neighbor_entry       *mpr_candidate=NULL;
   
  neighbor_hash_table=neighbortable.neighborhash;
  
  for (index=0;index<HASHSIZE;index++)
    
    {
      for(a_neighbor=neighbor_hash_table[index].neighbor_forw;a_neighbor!=(struct neighbor_entry  *)&neighbor_hash_table[index];a_neighbor=a_neighbor->neighbor_forw)
	{
	  if(maximum<a_neighbor->neighbor_2_nocov)
	    {
	      maximum=a_neighbor->neighbor_2_nocov;
	      mpr_candidate=a_neighbor;
	    }
	  
	}
    }
  return (mpr_candidate);
}


/*----------------------------------------------------------------------------*/
int
olsr_calculate_mpr()
     
{
  /*This function calculates the mpr neighbor*/
  
  olsr_u16_t                   two_hop_covered_count=0;
  olsr_u16_t                   two_hop_count=0;  
  struct neighbor_2_list_entry *two_hop_list=NULL;
  struct neighbor_2_list_entry *tmp;
  struct neighbor_entry        *mprs; 

  olsr_calculate_neighbors(&two_hop_count);
  two_hop_list=olsr_find_2_hop_neighbors_with_1_link();

  /*--------------------*/
  if(debug) {
    sprintf(auxstr,"Number of nodes: %d\r\n",two_hop_count);
	printf(auxstr);
  }
  /*--------------------*/

  while(two_hop_list!=NULL)
    {
      olsr_chosen_mpr(two_hop_list->neighbor_2->neighbor_2_nblist->neighbor,&two_hop_covered_count); 
      tmp=two_hop_list;
      two_hop_list=two_hop_list->neighbor_2_next;;
      free(tmp);
    }
  /*---------------------*/
  if(debug) {
    sprintf(auxstr,"Two hop nodes: %d\r\n",two_hop_covered_count);
	printf(auxstr);
  }
  /*---------------------*/

  while(two_hop_covered_count<two_hop_count)
    {
      mprs=olsr_find_maximum_covered();
      if (mprs!=NULL)
	{
	  olsr_chosen_mpr(mprs,&two_hop_covered_count);
	}
    }
  
  /*
    increment the mpr sequence number
  */
  neighbortable.neighbor_mpr_seq++;

	return 0;
}

/*-----------------------------------------------------------------------------*/
int
olsr_timed_out(struct timeval *timer)
{
  return( timercmp(timer,&now,<));
}
/*-----------------------------------------------------------------------------*/
int
olsr_init_timer(olsr_u32_t time_value,struct timeval *hold_timer)
{ 
  olsr_u16_t  time_value_sec=0;
  olsr_u16_t  time_value_msec=0;

  time_value_sec=time_value/1000;
  time_value_msec=time_value-(time_value_sec*1000);

  hold_timer->tv_sec=time_value_sec;
  hold_timer->tv_usec=time_value_msec*1000; 
  
	return 0;
}
/*-------------------------------------------------------------------------*/
void
olsr_linking_this_2_entries(struct neighbor_entry *neighbor,struct neighbor_2_entry *two_hop_neighbor)
{
  struct neighbor_list_entry    *list_of_1_neighbors;
  struct neighbor_2_list_entry  *list_of_2_neighbors;

  list_of_1_neighbors=(struct neighbor_list_entry *)malloc(sizeof(struct neighbor_list_entry));
  list_of_2_neighbors=(struct neighbor_2_list_entry *)malloc(sizeof(struct neighbor_2_list_entry));

  list_of_1_neighbors->neighbor=neighbor;
  list_of_1_neighbors->neighbor_next=two_hop_neighbor->neighbor_2_nblist;
  two_hop_neighbor->neighbor_2_nblist=list_of_1_neighbors;

  list_of_2_neighbors->neighbor_2=two_hop_neighbor;
  timeradd(&now,&hold_time_neighbor,&list_of_2_neighbors->neighbor_2_timer);

  list_of_2_neighbors->neighbor_2_next=neighbor->neighbor_2_list;
  neighbor->neighbor_2_list=list_of_2_neighbors;
  
  /*increment the pointer counter*/
  two_hop_neighbor->neighbor_2_pointer++;
}
/*-------------------------------------------------------------------------*/
int
olsr_process_message_neighbors(struct neighbor_entry *neighbor,struct hello_message *message)
{

  olsr_u8_t                    ok_insert=0;// boolean
  struct hello_neighbor        *message_neighbors;
  struct neighbor_entry        *in_neighbor;
  struct neighbor_2_list_entry *two_hop_neighbor_yet;
  struct neighbor_2_entry      *two_hop_neighbor;
  int count=0;

  message_neighbors=message->neighbors;
  
  while(message_neighbors!=NULL)
    {
      
      if(0!=memcmp(&message_neighbors->address,&addr_ip,sizeof(olsr_ip_addr)))	   	    
	if((message_neighbors->status==NBS_SYM)||(message_neighbors->status==NBS_MPR))
	  if((two_hop_neighbor_yet=olsr_lookup_my_neighbors(neighbor,&message_neighbors->address))!=NULL)
	    {
	      
	      /*
		Updating the holding time for this neighbor
	      */
	      timeradd(&now,&hold_time_neighbor,&two_hop_neighbor_yet->neighbor_2_timer);
	    
	      /*-------------------*/
		  if(debug) {
			sprintf(auxstr,"%s=============>Deja un voisin second niveau\r\n",convert_address_to_string(&message_neighbors->address));
			printf(auxstr);
		  }
	      /*-------------------*/
	   
	    }
	  else
	    {

	      /*-------------------*/
		  if(debug) {
			sprintf(auxstr,"%s=============>Un nouveau  voisin second niveau\r\n",convert_address_to_string(&message_neighbors->address));
			printf(auxstr);
		  }
	      /*-------------------*/

	      /*
		Looking if this node is already in the two hop neighbor table!!!
	      */
	      if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&message_neighbors->address))==NULL)
		{
		  ok_insert=0;
		  
		  /*
		    This node will be inserted in the two hop neighbor table even if it exist in the neighbor
		    table with the status ASYM or LOST. In this way we can have connectivity with this neighbor
		  */
		  
		  if((in_neighbor=olsr_lookup_neighbor_table(&message_neighbors->address))!=NULL)
		    {
		      if((in_neighbor->neighbor_status==NBS_ASYM)||(in_neighbor->neighbor_status==NBS_LOST))
			ok_insert=1;			      
		    }
		  else
		    ok_insert=1;
		  
		  if(ok_insert)
		    {
		      changes_neighborhood=UP;
		      changes_topology=UP;
		      two_hop_neighbor=(struct neighbor_2_entry *)malloc(sizeof(struct neighbor_2_entry));
		      two_hop_neighbor->neighbor_2_nblist=NULL;
		      two_hop_neighbor->neighbor_2_pointer=0;

		      memcpy(&two_hop_neighbor->neighbor_2_addr,&message_neighbors->address,sizeof(olsr_ip_addr));
		      olsr_insert_two_hop_neighbor_table(two_hop_neighbor);
		      olsr_linking_this_2_entries(neighbor,two_hop_neighbor);
		    }
		}
	      else
		{
		  
		/*
		  linking to this two_hop_neighbor entry
		*/	
		  olsr_linking_this_2_entries(neighbor,two_hop_neighbor); 
		}
	    }
      
     message_neighbors=message_neighbors->next;
      
    }
	return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_process_changes()
{
  /*------------------*/
  if(debug)
    {
      if(changes_neighborhood)
	printf("CHANGES IN NEIGHBORHOOD\r\n");
      if(changes_topology)
	printf("CHANGES IN TOPOLOGY\r\n");
    }
  /*------------------*/
  
  if(changes_neighborhood)
    {
      /*
	Calculate new mprs and routing table
       */
      olsr_calculate_mpr();
      olsr_calculate_routing_table();
      changes_neighborhood=DOWN;
      changes_topology=DOWN;
      return 0;
    }
  
  if(changes_topology)
    {
      /*
	calculate the routing table
      */
      olsr_calculate_routing_table();
      changes_topology=DOWN;
    }
  return 0;
}
/*-------------------------------------------------------------------------*/
int
olsr_process_received_hello(struct hello_message *message)
{
  /*
    
    1. If the entry already exists:
    
    1.1 the holding time of the entry is refreshed to
    NEIGHB_HOLD_TIME
    
    1.2 if the node finds its own address among the
    addresses listed in the HELLO message with a link type
    of ASYM_LINK, SYM_LINK or MPR_LINK, it updates
    the status of the link to the sender node as SYM_LINK if
    it was ASYM_LINK before.
    
    1.3 the N_2hop_list of the entry is updated to reflect the
    contents of the HELLO-message. 
    
    2. Otherwise a new entry is recorded in the Neighbor table with:
    
    2.1 N_addr is set to the address of the sender node
    
    2.2 N_status is set to the value of ASYM_LINK (asymmetric
    link)
    
    2.3 N_2hop_list is filled with the list of addresses
    contained in the HELLO message which have a link 
    type of SYM_LINK or MPR_LINK, and which are not 
    already present in the Neighbor table (i.e., they 
    are not one-hop neighbors). 
    
    If a node finds its own address in a HELLO message with
    a link type of ASYM_LINK, SYM_LINK or MPR_LINK, it does
    not register itself in the N_2hop_list. Rather, it
    changes the link status to the sender of the HELLO from
    ASYM_LINK to SYM_LINK. 
    
    2.4 N_time is set to the value of NEIGHB_HOLD_TIME
*/


/*==================================================================================*/
/*
  If upon recieving a HELLO message a node finds its own address in
  the the address list with a link type of "MPR", it must update the
  entry corresponding to the sender node's address in the MPR
  Selector table:
  
  1. If the entry already exists:
  
  1.1 if the MPR Sequence Number field of the HELLO message is
  greater than or equal to the MS_seq_num field of the entry
  in the table, then the MS_time is refreshed to NEIGHB_HOLD_TIME.
  
  1.2 if the MPR Sequence Number field of the HELLO message is
  greater than the MS_seq_num field of that entry, the
  MS_seq_num field is updated to the value of MPR Sequence
  Number field of the HELLO message.
  
  2. Otherwise, a new entry is recorded in the MPR Selector table,
  with:
  
  2.1 MS_addr is set to the address of sender of the HELLO
  message
  
  2.2 MS_seq_num is set to the MPR Sequence Number field of the
  HELLO message
  
  2.3 MS_time is set to the value of NEIGHB_HOLD_TIME
 
  
*/
  
  olsr_u8_t                 link=0;
  struct neighbor_entry     *neighbor;
  struct neighbor_2_entry   *two_hop_neighbor;
  

  if(message==NULL)
    return 1;

  link=olsr_lookup_link_status(message);

  if(link==NBS_MPR)
    olsr_update_mpr_selector_table(message);


    /*
      Mettre a jour la table des SMR
     */


  if(NULL== (neighbor=olsr_lookup_neighbor_table(&message->source_addr)))
     {

       neighbor=(struct neighbor_entry *)malloc(sizeof(struct neighbor_entry));       
       memcpy(&neighbor->neighbor_addr,&message->source_addr,sizeof(olsr_ip_addr));
       timeradd(&now,&hold_time_neighbor,&neighbor->neighbor_timer);
       
       neighbor->neighbor_ifp=NULL;
       neighbor->neighbor_2_list=NULL;
       neighbor->neighbor_2_nocov=0;
       
       olsr_insert_neighbor_table(neighbor);
       
       if((link==NBS_SYM)||(link==NBS_MPR)||(link==NBS_ASYM))
	 {
	   neighbor->neighbor_status=NBS_SYM;
	   /* This flag is set to UP to recalculate the MPR set and the routing table*/
	   changes_neighborhood=UP;
	   changes_topology=UP;
	   /*
	     remove this neighbor from the two hop table if exist
	   */
	   if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&neighbor->neighbor_addr))!=NULL)
	     olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
	 }
       else
	 neighbor->neighbor_status=NBS_ASYM;      
     }
  else
    {      

      timeradd(&now,&hold_time_neighbor,&neighbor->neighbor_timer);

      if(link!=NOT_DETECTED)
	{	  
	  switch (neighbor->neighbor_status)
	    {
	    case NBS_LOST:
	      neighbor->neighbor_status=NBS_ASYM;
	      break;
	    case NBS_ASYM:
	      if(link==NBS_LOST)
		 neighbor->neighbor_status=NBS_ASYM;
	      else
		{
		 neighbor->neighbor_status=NBS_SYM;
		 /* This flag is set to UP to recalculate the MPR set and the routing table*/
		 changes_neighborhood=UP;
		 changes_topology=UP;
		}
	      break;
	    case NBS_SYM:
	      if(link==NBS_LOST)
		{
		  neighbor->neighbor_status=NBS_ASYM;
		  /* This flag is set to UP to recalculate the MPR set and the routing table*/
		  changes_neighborhood=UP;
		  changes_topology=UP;
		}
	      break;
	    case NBS_MPR:
	      if(link==NBS_LOST)
		{
		  neighbor->neighbor_status=NBS_ASYM;
		  /* This flag is set to UP to recalculate the MPR set and the routing table*/
		  changes_neighborhood=UP;
		  changes_topology=UP;
		}
	      break;
	    default:;
	    }
	
	  /*
	    Deleting this neighbor from the two hop list if exist
	   */
	  if((neighbor->neighbor_status==NBS_MPR)||(neighbor->neighbor_status==NBS_SYM))
	    {
	      if((two_hop_neighbor=olsr_lookup_two_hop_neighbor_table(&neighbor->neighbor_addr))!=NULL)
		{
		  olsr_delete_two_hop_neighbor_table(two_hop_neighbor);
		  /* This flag is set to UP to recalculate the MPR set and the routing table*/
		  changes_neighborhood=UP;
		  changes_topology=UP;
		}
	    }
	}            
    }
  /*
    Processing the list of neighbors in the received message
  */
  olsr_process_message_neighbors(neighbor,message);

  olsr_destroy_hello_message(message);

  olsr_process_changes();

  /*---------- */
/*  
  if(debug)
    {
      olsr_print_neighbor_table();
      olsr_print_two_hop_neighbor_table();
      olsr_print_mpr_selector_table();
    }
*/
  /*---------*/

	return 0;
}

/*------------------------------------------------------------------------*/
int
olsr_process_received_tc(struct tc_message *message,struct tc_message **message_out)
{

  /* 
     1. If there exist some entry in the topology table whose T_last
     corresponds to the originator address of the TC message and
     whose T_seq is greater in value than the MSSN in the received
     message, then no further processing of this TC message is 
     performed and it is silently discarded (case: message
         received out of order).
	 
	 2. If there exist some entry in the topology table whose T_last
         corresponds to the originator address of the TC message and
         whose T_seq is lesser in value than the MSSN in the received
         message, then that topology entry is removed.
	 
	 3. For each of the MPR Selector address received in the TC
         message:
	 
         3.1 If there exist some entry in the topology table whose
	 T_dest corresponds to the MPR Selector address and the
	 T_last corresponds to the originator address of the TC
	 message, then the holding time T_time of that topology
	 entry is refreshed to TOP_HOLD_TIME.
	 
         3.2 Otherwise, a new topology entry is recorded in the
	 topology table whereas:
	 
	 - T_dest is set to the MPR Selector address,
	 - T_last is set to the originator address of the TC
	 message,
	 - T_seq is set to the value of MSSN received in the TC
	 message,
	 - T_time is set to the value of TOP_HOLD_TIME.
	 
	 
  */
  
  struct neighbor_entry           *neighbor;
  struct topology_last_entry      *t_last;


  struct tc_mpr_addr                *mpr;



  *message_out=NULL;
    
  if(message==NULL)
    return 1;

/*----------------------------*/
  if(debug)
    {
      mpr=message->multipoint_relay_selector_address;
      
      while(mpr!=NULL)
	{
	  sprintf(auxstr,"Start of Tc_received, the list of mprs in this message :This mpr is %s\r\n",convert_address_to_string(&mpr->address));
	  printf(auxstr);
	  mpr=mpr->next;
	}
    }
/*----------------------------*/

  if(0==memcmp(&message->originator,&addr_ip,sizeof(olsr_ip_addr)))
    {
      olsr_destroy_tc_message(message);
      return 1;
    }

  /*Check whether  the tc is received from a sym neighbor or not.
    Be careful we should not check the originator address*/

  if ((neighbor=olsr_lookup_neighbor_table(&message->source_addr))!=NULL)
    {
      if((neighbor->neighbor_status==NBS_ASYM)||(neighbor->neighbor_status==NBS_LOST))
	{
	  olsr_destroy_tc_message(message);
	  
	  return 1;
	}
    }
  else
    {

      /*The TC message source address is not in the neighbor table !!!!!!*/
      olsr_destroy_tc_message(message);
      return 1;
    }

  if(NULL!=olsr_lookup_duplicate_table(message))
    {
      
      olsr_destroy_tc_message(message);
      return 1;
    }

  olsr_insert_duplicate_table(message);

  t_last=olsr_lookup_last_topology_table(&message->originator);
  
  if(t_last!=NULL)
    {

      /*----------------------*/
      if(debug)
	{
	  printf("THE T_LAST EXISTS\r\n");
	}
      /*----------------------*/

      if(t_last->topology_seq>message->mssn)
	{
	  olsr_destroy_tc_message(message);
	  if(debug)
	    {
	      printf("TC_message TOO OLD\r\n");
	    }
	  return 1;
	}

      /*-----------------------*/
      if(debug)
	{
	  sprintf(auxstr,"THE NEW MSSN is %d and the old one is %d\r\n",message->mssn,t_last->topology_seq);
	  printf(auxstr);
	}
      /*-----------------------*/
      
      if(t_last->topology_seq<message->mssn)
	{
	  
	  olsr_delete_last_topolgy_table(t_last);
	  /*
	    We must insert the new entries contained in the received message
	   */ 
	  changes_topology=UP;
	  t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));
	  olsr_insert_last_topology_table(t_last,message);
	  olsr_update_last_table(t_last,message);
	}
      else
	{
	  /*
	    This the case where t_last->topology_seq==message->mssn, so
	    we must update the time out and create the new dest entries
	  */
	  
	  olsr_update_last_table(t_last,message);
	}
    }
  else
    {
      changes_topology=UP; /*changes_topology will be set to DOWN  after recalculating the routing table*/
      t_last=(struct topology_last_entry *)malloc(sizeof(struct topology_last_entry));
      olsr_insert_last_topology_table(t_last,message);
      olsr_update_last_table(t_last,message);
      
       
    }
  /*Check whether this node is a mpr of the previous one*/
  if(NULL!=(olsr_lookup_mpr_selector_table(&message->source_addr)))
    {
      /*Yes we are an mpr of the previous hop */
      /*set the source_addr to the current one*/
      /*---------------------*/
      if(debug) {
	sprintf(auxstr,".................WE R MPR OF %s\r\n",convert_address_to_string(&message->source_addr));
	printf(auxstr);
	  }
      /*---------------------*/

      memcpy(&message->source_addr,&addr_ip,sizeof(olsr_ip_addr));

      /*---------------------*/
      if(debug)
	{
	  sprintf(auxstr,"........::.THE ORIGINATOR OF THIS MPR %s\r\n",convert_address_to_string(&message->originator));
	  printf(auxstr);
	  sprintf(auxstr,"...NOW WE R  THE SOURCE OF THIS MSG %s\r\n",convert_address_to_string(&message->source_addr));
	  printf(auxstr);
	}
      /*---------------------*/
      
      /*send this message */
      
      //IMPORTANT Don't forget to decrement the hop count in the message!!!!!!!
      /*--------------------*/
      if(debug) {
	sprintf(auxstr,"The hop count is %d\r\n",message->hop_count);
	printf(auxstr);
	  }
      /*--------------------*/
      
       message->hop_count --;
             
       if(message->hop_count>0)
	 {
	   *message_out=message;
	 }
       else
	 olsr_destroy_tc_message(message);
       
    }
  else
      olsr_destroy_tc_message(message);      
  /*--------------- */    
  if(debug)
    olsr_print_topology_table();
  /*---------------*/
  olsr_process_changes();
  
  return 0; 
}

/*------------------------------------------------------------------------*/
int
olsr_delete_time_out_entries()
{
 
  olsr_time_out_duplicate_table();
  olsr_time_out_mpr_selector_table();
  olsr_time_out_neighborhood_tables();
  olsr_time_out_topology_table();
  
  olsr_process_changes();
  
	return 0;  
}
/*------------------------------------------------------------------------*/
void
olsr_release_tables(int number)
{
  /*
    neighbortable
    neighbor2table
    duplicatetable
    mprstable
    routingtable
    topologytable
  */

  olsr_release_duplicate_table();
  olsr_release_mpr_seletector_table();
  olsr_release_neighbor_table();

  olsr_release_routing_table(mirror_table);
  olsr_release_routing_table(routingtable);

  olsr_release_topology_table();
  olsr_release_two_hop_neighbor_table();
}
/*------------------------------------------------------------------------*/
int
olsr_init_tables()
{

  /*   
    neighbortable
    neighbor2table
    duplicatetable
    mprstable
    routingtable
    topologytable
    topologylasttable
  */

  olsr_16_t index;

  /*-------------------*/
  olsr_init_trace_route_file();
  /*-------------------*/

  olsr_init_timer(NEIGHB_HOLD_TIME*1000,&hold_time_neighbor);
  olsr_init_timer(TOP_HOLD_TIME*1000,&hold_time_topology);
  olsr_init_timer(DUPLICATE_HOLD_TIME*1000,&hold_time_duplicate_tc);

  neighbortable.neighborhash =(struct neighborhash *) calloc (HASHSIZE,sizeof(struct neighborhash));
  mprstable.mpr_selector_hash=(struct mpr_selector_hash *) calloc(HASHSIZE,sizeof(struct mpr_selector_hash));
  
  for(index=0;index<HASHSIZE;index++)
    {
      neighbortable.neighborhash[index].neighbor_forw=(struct	neighbor_entry *)&neighbortable.neighborhash[index];
      neighbortable.neighborhash[index].neighbor_back=(struct	neighbor_entry *)&neighbortable.neighborhash[index];
      neighbortable.neighbor_mpr_seq=0;
      
      neighbor2table[index].neighbor2_forw=(struct neighbor_2_entry *)&neighbor2table[index];
      neighbor2table[index].neighbor2_back=(struct neighbor_2_entry *)&neighbor2table[index];

      duplicatetable[index].duplicate_forw=(struct duplicate_entry *)&duplicatetable[index];
      duplicatetable[index].duplicate_back=(struct duplicate_entry *)&duplicatetable[index];

      mprstable.mpr_selector_hash[index].mpr_selector_forw=(struct mpr_selector_entry *)&mprstable.mpr_selector_hash[index];
      mprstable.mpr_selector_hash[index].mpr_selector_back=(struct mpr_selector_entry *)&mprstable.mpr_selector_hash[index];
      mprstable.mssn=0;

      routingtable[index].rt_forw=(struct rt_entry *)&routingtable[index];
      routingtable[index].rt_back=(struct rt_entry *)&routingtable[index];

      mirror_table[index].rt_forw=(struct rt_entry *)&mirror_table[index];
      mirror_table[index].rt_back=(struct rt_entry *)&mirror_table[index];

      topologytable[index].topology_destination_forw=(struct topology_destination_entry *)&topologytable[index];
      topologytable[index].topology_destination_back=(struct topology_destination_entry *)&topologytable[index];

      topologylasttable[index].topology_last_forw=(struct topology_last_entry *)&topologylasttable[index];
      topologylasttable[index].topology_last_back=(struct topology_last_entry *)&topologylasttable[index];
    }

  
	return 0;  
}
/*------------------------------------------------------------------------*/
int
olsr_build_hello_packet(struct hello_message *message)
{
  struct hello_neighbor *message_neighbor;
  struct neighbor_entry *neighbor;
  struct neighborhash   *hash_neighbor;
  olsr_u16_t             index;

  message->neighbors=NULL;
  message->packet_seq_number=0;

  message->mpr_seq_number=neighbortable.neighbor_mpr_seq;

  /*-----------------*/
  if(debug) {
    sprintf(auxstr,"mpr is %d\r\n",message->mpr_seq_number);
	printf(auxstr);
  }
  /*-----------------*/

  memcpy(&message->source_addr,&addr_ip,sizeof(olsr_ip_addr));

 
  for(index=0;index<HASHSIZE;index++)
    {       
      hash_neighbor=&neighbortable.neighborhash[index];
      
      for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	{	    
	  message_neighbor=(struct hello_neighbor *)malloc(sizeof(struct hello_neighbor));
	  message_neighbor->status=neighbor->neighbor_status;
	  memcpy(&message_neighbor->address,&neighbor->neighbor_addr,sizeof(olsr_ip_addr));

	  /*--------------------*/
	  if(debug)
	    {
	      sprintf(auxstr,"%s           \r\n ",convert_address_to_string(&message_neighbor->address));
		  printf(auxstr);
	      sprintf(auxstr," status  %d\r\n",message_neighbor->status);
		  printf(auxstr);
	    }
	  /*--------------------*/

	  message_neighbor->next=message->neighbors;
	  message->neighbors=message_neighbor;	    
	}
    }
  return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_build_tc_packet(struct tc_message *message)
{
  struct tc_mpr_addr        *message_mpr;
  struct mpr_selector_entry *mprs;
  struct mpr_selector_hash  *mprs_hash;
  olsr_u16_t          index;

  message->multipoint_relay_selector_address=NULL;
  message->packet_seq_number=0;
 
  message->hop_count=HOPCNT_MAX;
  message->mssn=mprstable.mssn;

  memcpy(&message->originator,&addr_ip,sizeof(olsr_ip_addr));
  memcpy(&message->source_addr,&addr_ip,sizeof(olsr_ip_addr));
  
  for(index=0;index<HASHSIZE;index++)
    {
      mprs_hash=&mprstable.mpr_selector_hash[index];
      
      for(mprs=mprs_hash->mpr_selector_forw;mprs!=(struct mpr_selector_entry *)mprs_hash;mprs=mprs->mpr_selector_forw)
	{
	  message_mpr=(struct tc_mpr_addr *)malloc(sizeof(struct tc_mpr_addr));
	  memcpy(&message_mpr->address,&mprs->mpr_selector_addr,sizeof(olsr_ip_addr));
	  message_mpr->next=message->multipoint_relay_selector_address;
	  message->multipoint_relay_selector_address=message_mpr;
	}      
    }
  return 0;
}


/*------------------------------------------------------------------------*/

int
olsr_build_routing_packet(struct rt_message **message)
{

  struct rt_entry     *destination;
  struct rthash       *routing_hash;
  olsr_u8_t           index;
  struct rt_message   *routing_message=NULL;

  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&routingtable[index];
      for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
      {
        routing_message= (struct rt_message *) malloc(sizeof(struct rt_message));
	memcpy(&routing_message->destination,&destination->rt_dst,sizeof(olsr_ip_addr));

	if(debug) {
	  sprintf(auxstr,"%s            \r\n",convert_address_to_string(&destination->rt_dst));
	  printf(auxstr);
	}

        memcpy(&routing_message->next_hop,&destination->rt_router,sizeof(olsr_ip_addr));

		if(debug) {
		sprintf(auxstr,"%s              \r\n",convert_address_to_string(&destination->rt_router));
		printf(auxstr);
		}

        routing_message->hop_count=destination->rt_metric;
	
        routing_message->next=*message;
        *message=routing_message;
	if (debug)
	  printf("Working in ROUTE.\n");

      }
   }

  return(0);
}

/*------------------------------------------------------------------------*/
int
olsr_print_tables()
{
  olsr_print_duplicate_table();
  olsr_print_mpr_selector_table();
  olsr_print_neighbor_table();
  olsr_print_two_hop_neighbor_table();
  olsr_print_routing_table(routingtable);
  olsr_print_topology_table();

	return 0;
}
/*------------------------------------------------------------------------*/
char 
*convert_sockaddr_to_string(struct sockaddr *address_to_convert)
{
  struct sockaddr_in           *address;
  
  address=(struct sockaddr_in *)address_to_convert; 
  return(inet_ntoa(address->sin_addr));
  
}
/*------------------------------------------------------------------------*/
char 
*convert_address_to_string(olsr_ip_addr *address)
{

  struct in_addr in;
  in.s_addr=*address;
  return(inet_ntoa(in));
  
}
/*--------------------------------------------------------------------*/
int
olsr_init_trace_route_file()
{

 if ((f_r_id=fopen("routes","w"))==NULL)
    {
      printf("error\r\n");
      exit(0);
    }

	return 0;
}














