/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 *	$Id: olsr.h,v 1.2 2000/12/06 10:36:11 prima Exp $
 */

/*
 * Routing Table Management Daemon
 */

#ifndef _PROTOCOLS_OLSRD_H
#define	_PROTOCOLS_OLSRD_H 1



/*
 * Optimized Link State Routing Protocol
 */

/*
 * Hello packet informations
 */

struct hellinfo {
        olsr_u8_t link_type;                /* link type */
        olsr_u8_t res[1];                   /* padding   */
        olsr_u16_t  next_linktype;         /* offset to the next link type */
        olsr_ip_addr neigh_addr[1];	    /* neighbor address 
	        			    (variable length)*/
};

struct hellomsg {
        olsr_u16_t  hell_seqnum;              /* msg sequence number  */
        olsr_u16_t  hell_mprseqnum;           /* mpr sequence number  */
	struct hellinfo hell_info[1];         /* variable length... */
};
 
/*
 * Topology Control packet informations
 */

struct tcmsg {
        olsr_u16_t  tc_seqnum;              /* msg sequence number  */
        olsr_u16_t  tc_mprseqnum;           /* mprs sequence number  */
        olsr_u8_t  tc_hopcnt;              /* maximum number of hops  */
        olsr_u8_t tc_res[3];              /* pad to 32-bit boundary   */
        olsr_ip_addr tc_origaddr;	 /* originator address */
        olsr_ip_addr tc_mprsaddr[1];	 /* mprs address 
	        			    (variable length)*/
};

/*
 * ROUTE_REQUEST response packet informations
 */

struct rtmsg {
        olsr_u8_t  rt_hopcnt;           /* number of hops to destination */
        olsr_u8_t rt_res[3];            /* pad to 32-bit boundary   */
        olsr_ip_addr rt_destaddr;       /* destination address */
        olsr_ip_addr rt_nextaddr;	/* next hop address */
};

/*
 * OLSR packet
 */

struct olsrmsg {
        olsr_u8_t olsr_msgtype;            /* msg type */
        olsr_u8_t olsr_broad;              /* msg to broadcast? */
        olsr_u16_t  olsr_nextmsg;            /* next link msg */
        union {
		struct	hellomsg un_hello[1]; /* variable length... */
		struct	tcmsg un_tc[1];	      /* variable length... */
	        struct  rtmsg un_rt[1];       /* response to ROUTE_REQUEST */
	        char    un_tracefile[1];      /* trace file for trace 
						 fonctioonalities */
	} olsrun;
#define	olsr_hello	olsrun.un_hello
#define	olsr_tc	        olsrun.un_tc
#define	olsr_rt 	olsrun.un_rt
#define	olsr_tracefile 	olsrun.un_tracefile
};


struct olsr {
	olsr_u16_t	olsr_packlen;		/* packet length */
	olsr_u8_t	olsr_res[2];		/* pad to 32-bit boundary */
        struct olsrmsg olsr_msg[1];
};
 

/*
 * Message types.
 */

#define	HELLO_PACKET		1     /* neighborhood  info */
#define	TC_PACKET		2     /* topology information */
#define	HELLO_REQUEST		3     /* to obtain neighborhood  info */
#define	ROUTE_REQUEST		4     /* to obtain routing table contents */
#define	ROUTE_RESPONSE		5     /* routing table information */
#define	OLSRD_TRACEON		6     /* to enable trace fonctionnalities */
#define	OLSRD_TRACEOFF		7     /* to desable trace fonctionnalities */
#define	UPDATE_NEIGHBOR		8     /* to update the neighborhood  */

#define	OLSRTYPE_MAX		9

#ifdef OLSRTYPES
char *olsrtypes[OLSRTYPE_MAX] =
  { "#0", "HELLO MSG", "TC MSG", "HELLO REQUEST", "ROUTE REQUEST",
    "ROUTE RESPONSE", "TRACE ON MSG", "TRACE OFF MSG", 
    "UPDATE NEIGHBORHOOD MSG" };
#endif

#define	MSG_NOBROADCAST		0	/* don't broadcast the msg */
#define	MSG_BROADCAST		1	/* msg to broadcast */

#define	HOPCNT_MAX		16	/* maximum hops number */
#define	MAXPACKETSIZE		512	/* max broadcast size */

/*
 * Timer values used in managing the routing table.
 * Neighbor tables are broadcast every HELLO_INTERVAL seconds and topology 
 * tables every TC_INTERVAL.
 * If changes occur in the MPR Selector set, the next TC information is 
 * sent after TC_MIN_INTERVAL, time starting from the time the last TC 
 * message was sent.
 * Every update of a neighbor or topology entry forces an entry's timer 
 * to be reset.
 * After .._HOLD_TIME without updates,the entry is delete.
 */

#define	TIMER_RATE      2   /* alarm clocks every 2 seconds */

#define	HELLO_INTERVAL  2  /* time to supply neighbor database */
#define	TC_INTERVAL     6   /* time to supply topology informations */
#define	MIN_TC_INTERVAL	2   /* min. interval to broadcast topology changes */
#define	MAX_TC_INTERVAL	MIN_TC_INTERVAL   /* max. time to delay topology changes */


#define	NEIGHB_HOLD_TIME     6  /* time to delete entry in nieghbor table */
#define	TOP_HOLD_TIME	    16 /* time to delete entry in topology table */
#define DUPLICATE_HOLD_TIME (2 * TC_INTERVAL) 

#endif /* protocols/olsrd.h */



struct neighbor_2_list_entry 
*olsr_find_2_hop_neighbors_with_1_link();

int
olsr_chosen_mpr(struct neighbor_entry *one_hop_neighbor,olsr_u16_t *two_hop_covered_count);

int
olsr_calculate_neighbors(olsr_u16_t *two_hop_count);

struct neighbor_entry 
*olsr_find_maximum_covered();

int
olsr_calculate_mpr();

int
olsr_timed_out(struct timeval *timer);

int
olsr_init_timer(olsr_u32_t time_value,struct timeval *hold_timer);

void
olsr_linking_this_2_entries(struct neighbor_entry *neighbor,struct neighbor_2_entry *two_hop_neighbor);

int
olsr_process_message_neighbors(struct neighbor_entry *neighbor,struct hello_message *message);

int
olsr_process_received_hello(struct hello_message *message);

int
olsr_process_received_tc(struct tc_message *message,struct tc_message **message_out);

int
olsr_delete_time_out_entries();

void
olsr_release_tables(int number);

int
olsr_init_tables();

int
olsr_build_hello_packet(struct hello_message *message);

int
olsr_build_tc_packet(struct tc_message *message);

int
olsr_build_routing_packet(struct rt_message **message);

int
olsr_print_tables();

char 
*convert_sockaddr_to_string(struct sockaddr *address_to_convert);

int
olsr_hashing(olsr_ip_addr *address,olsr_u32_t *h);

char 
*convert_address_to_string(olsr_ip_addr *address);

int
olsr_init_trace_route_file();
