/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

#include "stdafx.h"

char routing_table_rcsid[] = 
  "$Id: routing_table.c,v 1.2 2000/12/06 10:36:12 prima Exp $";


#include "defs.h"

struct rthash                     routingtable[HASHSIZE];

/*------------------------------------------------------------------------*/
struct rt_entry
*olsr_lookup_routing_table(olsr_ip_addr *dst)
{

  struct rt_entry *destination;
  struct rthash   *routing_hash;
  olsr_u32_t      hash;

  olsr_hashing(dst,&hash);
  routing_hash=&routingtable[hash & HASHMASK];
  
  for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
    {
      if(destination->rt_hash!=hash)
	continue;

      if (0==memcmp(&destination->rt_dst,dst,sizeof(olsr_ip_addr)))
	return(destination);
    }
  return(NULL);
  
}
/*------------------------------------------------------------------------*/
int
olsr_delete_routing_table(struct rt_entry* destination)
{

  if(destination->rt_ifp!=NULL)
    free(destination->rt_ifp);    
  olsr_remque((struct olsr_qelem *)destination);
  free((void *)destination);

	return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_release_routing_table(struct rthash *table)
{ 
  struct rt_entry *destination;
  struct rt_entry *destination_tmp;
  struct rthash   *routing_hash;
  olsr_u8_t             index;
  

  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&table[index];
      destination=routing_hash->rt_forw;
      while(destination!=(struct rt_entry *)routing_hash)
	{
	  destination_tmp=destination;
	  destination=destination->rt_forw;
	  olsr_delete_routing_table(destination_tmp);
	}
    }

	return 0;
}
/*------------------------------------------------------------------------*/
struct rt_entry 
*olsr_insert_routing_table(olsr_ip_addr *dst)
{
  struct rt_entry *new_route_entry;
  struct rthash   *routing_hash;
  olsr_u32_t       hash;

  olsr_hashing(dst,&hash);
  routing_hash=&routingtable[hash & HASHMASK];

  new_route_entry=(struct rt_entry *)malloc(sizeof(struct rt_entry));
  new_route_entry->rt_hash=hash;

  memcpy(&new_route_entry->rt_dst,dst,sizeof(olsr_ip_addr));

  /*
    No timeout is needed since the routing table is calculated for each network change
   */

  new_route_entry->rt_metric=0;
  new_route_entry->rt_ifmetric=0;
  new_route_entry->rt_state=0; 
  new_route_entry->rt_flags=(RTF_UP|RTF_GATEWAY|RTF_HOST);
  new_route_entry->rt_ifp=NULL;

  olsr_insque((struct olsr_qelem *)new_route_entry, (struct olsr_qelem *)routing_hash);
  
  return(new_route_entry);
  /*
    We need this pointer to use it for the next routing calculation
   */
}
/*------------------------------------------------------------------------*/
struct rt_entry 
*olsr_insert_routing_table_from_topo(struct rt_entry *r_last,struct topology_destination_entry *destination)
{
  struct rt_entry *new_route_entry=NULL;
  struct rthash   *routing_hash;
  olsr_u32_t      hash;

  /*
    the hash calculation is the same
   */
  /*-----------------*/
  if (debug)
    printf("Insertion from the topology/////////////////\n");
  /*-----------------*/
  new_route_entry=olsr_insert_routing_table(&destination->topology_destination_dst);
  new_route_entry->rt_flags=(RTF_UP|RTF_HOST|RTF_GATEWAY);
  memcpy(&new_route_entry->rt_router,&r_last->rt_router,sizeof(olsr_ip_addr));
  new_route_entry->rt_metric=r_last->rt_metric+1; 
   
 return(new_route_entry);
  
}
/*------------------------------------------------------------------------*/
int
olsr_add_default_gateway()
{
  /*
    This function adds a new entry for the default destination
  */
  struct rt_entry *default_gateway_entry;
  struct rt_entry  *gateway_entry;
  olsr_ip_addr     inaddr_any=INADDR_ANY;
  
  if(default_gateway!=0)
    
    if(NULL!=(gateway_entry=olsr_lookup_routing_table(&default_gateway)))
      {      
	default_gateway_entry=olsr_insert_routing_table(&inaddr_any); 
	default_gateway_entry->rt_flags=0;
	default_gateway_entry->rt_flags=(RTF_UP|RTF_GATEWAY);
	memcpy(&default_gateway_entry->rt_router,&gateway_entry->rt_router,sizeof(olsr_ip_addr));
	default_gateway_entry->rt_metric=gateway_entry->rt_metric;
      }

	return 0;
}
/*------------------------------------------------------------------------*/
struct destination_n
*olsr_fill_routing_table_with_neighbors()
{
  struct destination_n   *list_destination_n=NULL;
  struct destination_n   *list_destination_tmp=NULL;
  olsr_u8_t              index;
  struct neighbor_entry  *neighbor;
  struct neighborhash    *hash_neighbor;
  struct rt_entry        *new_route_entry=NULL;
 
  for(index=0;index<HASHSIZE;index++)
    {
      hash_neighbor=&neighbortable.neighborhash[index];
      for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)     
	{
	  if((neighbor->neighbor_status==NBS_SYM)||(neighbor->neighbor_status==NBS_MPR))
	    {
	      new_route_entry=olsr_insert_routing_table(&neighbor->neighbor_addr);
	      memcpy(&new_route_entry->rt_router,&neighbor->neighbor_addr,sizeof(olsr_ip_addr));
	      new_route_entry->rt_flags=(RTF_UP|RTF_HOST);
	      /*The next router is the neighbor itself*/
	      new_route_entry->rt_metric=1; 
	      /*fisrt hop. may be we should use another metric*/
	      list_destination_tmp=(struct destination_n *)malloc(sizeof(struct destination_n));
	      list_destination_tmp->destination=new_route_entry;
	      list_destination_tmp->next=list_destination_n;
	      list_destination_n=list_destination_tmp;
	      
	    }
	}
    }
  return(list_destination_n);
}
/*------------------------------------------------------------------------*/
int 
olsr_calculate_routing_table()

{
  struct destination_n               *list_destination_n=NULL;
  struct destination_n               *list_destination_n_1=NULL;
  struct destination_n               *destination_n_1=NULL;
  struct topology_last_entry         *topo_last;
  struct  destination_list           *topo_dest;
  struct rt_entry                    *destination_entry;
  
  olsr_routing_mirror();

  /*--------------------------*/
  if(debug)
    {
      printf(">>>>State of the tables after mirror copy\r\n");
      printf("The new routing table is: \r\n");
      olsr_print_routing_table(routingtable);
      printf("The old routing table is:\r\n");
      olsr_print_routing_table(mirror_table);
      printf("**************************************\n");
    }
  /*--------------------------*/

  /*
    Delete old entries from the table
  */
  list_destination_n_1=olsr_fill_routing_table_with_neighbors();
  list_destination_n=list_destination_n_1;
  
  while(list_destination_n_1!=NULL)
    {
      
      list_destination_n_1=NULL;
      while(list_destination_n!=NULL)
	
	{
	  /*----------------------*/
	  if (debug)
	    printf("%s+++++++++++ last hop\n",convert_address_to_string(&list_destination_n->destination->rt_dst));
	  /*----------------------*/
	  if(NULL!=(topo_last=olsr_lookup_last_topology_table(&list_destination_n->destination->rt_dst)))
	    {
	      topo_dest=topo_last->topology_list_of_destinations;
	      while(topo_dest!=NULL)
		{
		  /*
		    topo_dest is not in the routing table		   
		  */
		  if(NULL==olsr_lookup_routing_table(&topo_dest->destination_node->topology_destination_dst))
		    
		    {
		      destination_n_1=(struct destination_n *)malloc(sizeof(struct destination_n));
		      destination_n_1->destination=olsr_insert_routing_table_from_topo(list_destination_n->destination,topo_dest->destination_node);

		      destination_n_1->next=list_destination_n_1;
		      list_destination_n_1=destination_n_1;
		      
		    }
		  topo_dest=topo_dest->next;
		  
		}
	    }
	  
	  
	  destination_n_1=list_destination_n;
	  list_destination_n=list_destination_n->next;
	  free(destination_n_1);
	  	  
	}
      list_destination_n=list_destination_n_1;
    }
  
  olsr_add_default_gateway();
  
  olsr_update_kernel_routes();

  /*-----------------*/
  if(debug)
    {
      olsr_print_routing_table(routingtable);
      printf("The old routing table is:\r\n");
      olsr_print_routing_table(mirror_table);
    }
  /*-----------------*/

  olsr_release_routing_table(mirror_table);

	return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_print_trace_routing_table(struct rthash *table)
{

  struct rt_entry     *destination;
  struct rthash       *routing_hash;
  olsr_u8_t           index;
  char * timestr;

  //fprintf(f_r_id,"Current time is %s\n",ctime((time_t *)&now.tv_sec));
  PICAgetTimeStr(&timestr);
  fprintf(f_r_id,"Current time is %s\r\n",timestr);
  fprintf(f_r_id,"ROUTING TABLE\r\n");
  fprintf(f_r_id,"DESTINATION       NEXT HOP\r\n");
  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&table[index];
      for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
	{

	  fprintf(f_r_id,"%s            ",convert_address_to_string(&destination->rt_dst));
	  fprintf(f_r_id,"%s           \n ",convert_address_to_string(&destination->rt_router));
	}
    }
  fflush(f_r_id);

	return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_print_routing_table(struct rthash *table)
{

  struct rt_entry     *destination;
  struct rthash       *routing_hash;
  olsr_u8_t           index;

  printf("ROUTING TABLE\r\n");
  printf("DESTINATION       NEXT HOP\r\n");
  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&table[index];
      for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
	{
	  printf("%s            ",convert_address_to_string(&destination->rt_dst));
	  printf("%s           \r\n ",convert_address_to_string(&destination->rt_router));
	}
    }

	return 0;
}
/*------------------------------------------------------------------------*/
